#include <stdio.h>
#include <stdlib.h>     // for ?
#include <string.h>     // for memXXX
//#include <unistd.h>     // for ?
#include <io.h>

#include <sys/types.h>  // for open
#include <sys/stat.h>
#include <fcntl.h>

#if 0       //DEBUG
 #define DEBUG
 #define inline
 #define static
 #define DPRINTF(a) printf a
#else       // not DEBUG
 #define NDEBUG
 #define DPRINTF(a)
#endif

typedef unsigned long  u32; typedef signed long  s32;
typedef unsigned short u16; typedef signed short s16;
typedef unsigned char  u8;  typedef signed char  s8;

const u16 _16k = 16*1024u;
const u16 _32k = 32*1024u;
const u32 _64k = 64*1024ul;
const u32 _1m = 1024ul*1024ul;

static int die(const char* msg) {
    puts(msg);
    exit(1);
    return 0;
}

static void read_or_die(int fd, void *buf, u16 size, const char* msg) {
    u16 cnt = read(fd, buf, size);
    (cnt == size) || die(msg);
}

static u8* malloc_or_die(u16 size, const char* msg) {
    void* p = malloc(size);
    p || die(msg);
    return (u8*)p;
}

inline u16 seg(const void far* p) { return u32(p)>>16; }
inline u16 ofs(const void* p) { return u16(u32(p)); }
//inline u32 ptr2lin(const void* p) { return (u32(seg(p))<<4) + ofs(p); }

//GCC:
//inline void barrier()  { __asm__ __volatile__("": : :"memory"); }
//inline void cli() { __asm__ __volatile__("cli"); }
//inline void sti() { __asm__ __volatile__("sti"); }

static void barrier() {}
static void cli() { asm { cli } }
static void sti() { asm { sti } }

//DEBUG: shows screen full of hearts and freeze
static void hang() {
    asm {
        cli
        cld
        mov     cx,0b800h
        mov     es,cx
        xor     di,di
        mov     al,03h
        rep stosb
    }
z:  goto z;
}

static int check_a20() {
    volatile u8 far* addr0 = (volatile u8 far*)0x00000000L;
    volatile u8 far* addr1 = (volatile u8 far*)0xFFFF0010L;

    cli();
    u8 sv0 = *addr0;
    u8 sv1 = *addr1;
    int ok=0;
    *addr0=0x55;
    barrier();
    *addr1=0xAA;
    barrier();
    if(*addr0==0x55) ok=1;
    *addr0=sv0;
    *addr1=sv1;
    sti();

    return ok;
}

// Asm helpers
extern "C" int far is_rm32();
extern "C" void far memcpy32(u16,u32, u16,u32, u32);
extern "C" void far set_sregs_jump_seg_ofs(u16 sr, u16 sp, u16 seg, u16 ofs);
extern "C" u32 far xmm_driver();
extern "C" u32 far xmm_alloc(u32 drv, u32 size);
extern "C" u32 far memtop_e801();   // TODO: for initrd
extern "C" u32 far memtop_88();

/***************
    Memory layout assumed by kernel boot process
    --------------------------------------------
    Note: claims that kernel setup is relocatable are
    still not 100% valid.
    bzImage decompressing trashes 10000-8ffff range,
    so rm loader -> pm kernel info is lost if it was here...
    So I had to stick to 90000.

10000000+------------------------+  <- 256m
        |  initrd                |      TODO: initrd is at top of mem, but
        |                        |      not higher than 256m
        +------------------------+
        |  bzImage               |      bzImage is at 1m
        |                        |      XMS or 64k offset tricks used...
00100000+------------------------+  <- 1m
        |  video, BIOS etc       |      Do not use.
000A0000+------------------------+
        |  Reserved for BIOS     |      Do not use.  Reserved for BIOS EBDA.
0009A000+------------------------+  <- stack top for kernel rm code
        |  Cmdline               |
00098000+------------------------+  <- heap top for kernel rm code
        |  Kernel setup          |      The kernel real-mode code.
00090200+------------------------+
        |  Kernel boot sector    |      The kernel legacy boot sector.
00090000+------------------------+
        |  Zapped by ungzip      |      Historically zImages were loaded here
        | (zImage once was here) |      bzImages use this space for ungzipping
00010000+------------------------+
        |  Boot loader           |  <- Boot sector entry point 0000:7C00
00001000+------------------------+
        |  Reserved for MBR/BIOS |
00000800+------------------------+
        |  Typically used by MBR |
00000600+------------------------+
        |  BIOS use only         |
00000000+------------------------+
*/

const char* image_name = "bzImage";
const char* initrd_name = 0;   //TODO: "initrd";
const char* cmdline = "auto";
u16 vid_mode=0x303;     // -3 = ask
                        // -2 = Extended VGA
                        // -1 = Normal VGA
                        //  n = as "n" was pressed
u8* rm_buf;
u16 rm_size;
u32 pm_buf;
u32 pm_size;

struct first1k_t {
    u8      pad0[0x01f1];
    u8      setup_sects;    // 01F1 The size of the setup in sectors
    u16     root_flags;     // 01F2 If set, the root is mounted readonly
    u16     syssize;        // 01F4 DO NOT USE - for bootsect.S use only
    u16     swap_dev;       // 01F6 DO NOT USE - obsolete
    u16     ram_size;       // 01F8 DO NOT USE - for bootsect.S use only
    u16     vid_mode;       // 01FA Video mode control
    u16     root_dev;       // 01FC Default root device number
    u16     boot_flag;      // 01FE 0xAA55 magic number
    u16     jump;           // 0200 Jump instruction
    u32     header;         // 0202 Magic signature "HdrS"
    u16     version;        // 0206 Boot protocol version supported
    u32     realmode_swtch; // 0208 Hook called just before rm->pm
    u32     start_sys;      // 020C Points to kernel version string
    u8      type_of_loader; // 0210 Boot loader identifier
    u8      loadflags;      // 0211 Boot protocol option flags
    u16     setup_move_size;// 0212 Move to high memory size (used with hooks)
    u32     code32_start;   // 0214 Boot loader hook (see below)
    u32     ramdisk_image;  // 0218 initrd load address (set by boot loader)
    u32     ramdisk_size;   // 021C initrd size (set by boot loader)
    u32     bootsect_kludge;// 0220 DO NOT USE - for bootsect.S use only
    u16     heap_end_ptr;   // 0224 Free memory after setup end
    u16     pad1;           // 0226 Unused
    u32     cmd_line_ptr;   // 0228 32-bit pointer to the kernel command line
    u8      pad2[0x400-0x22c];  // 022C
}; //__attribute((packed));

const u32 HdrS = 'H' + ('d'<<8) + (u32('r')<<16) + (u32('S')<<24);

static void _loadds _saveregs far last_ditch() {
    cli();
    // Move kernel
    memcpy32(
        0, _1m,         // dst seg,ofs
        0, pm_buf,      // src seg,ofs
        pm_size         // size
    );
    // TODO: Move initrd
}

void syntax() {
    die("Syntax:\n"
        "LINLD [image=<filename>] [initrd=<filename>] [cl=<kernel command line>]\n"
        "Defaults:\n"
        "    image=bzImage\n"
        "    initrd=(nothing)\n"
        "    cl=auto\n"
        "Use quotes: \"cl=...\" if you want to embed spaces into cmdline"
    );
}

int main(int argc, const char* argv[]) {
    puts("LINLD v0.2");

#if sizeof(first1k_t)!=0x400
    die("BUG: Bad first1k");
#endif
#ifdef DEBUG
    printf("Memory top according to int15:E801: %08lx\n", memtop_e801());
    printf("Memory top according to int15:88: %08lx\n", memtop_88());
    printf("XMM driver entry point: %08lx\n", (u32)xmm_driver());
    printf("argc=%d\n", argc);
    {for(int i=0;i<argc;i++) {
        printf(" argv[%d]=%s\n", i, argv[i]);
    }}
    //exit(1);
#endif

    // Parse command line

    if(argc<2) syntax();
    {for(int i=1;i<argc;i++) {
        if(strncmp(argv[i],"image=",6) == 0) {
            image_name=argv[i]+6;
        }
        else if(strncmp(argv[i],"initrd=",7) == 0) {
            initrd_name=argv[i]+7;
        }
        else if(strncmp(argv[i],"cl=",3) == 0) {
            cmdline=argv[i]+3;
        }
        //else if(strncmp(argv[i],"vga=",4) == 0) { ... }
        else
            syntax();
    }}
    DPRINTF(("image_name  ='%s'\n",image_name ));
    DPRINTF(("initrd_name ='%s'\n",initrd_name));
    DPRINTF(("cmdline     ='%s'\n",cmdline    ));
    //exit(1);
    if(initrd_name) die("Sorry, initrd isn't supported yet");

    // Check environment

    is_rm32() || die("I need 386+ CPU running in real mode");
    //TODO: do not give up, enable if not enabled yet
    check_a20() || die("Address line A20 isn't enabled");

    // Open kernel, read first kb, check it

    rm_buf = malloc_or_die(_32k, "Can't allocate rm buf");
    int fd = open(image_name, O_RDONLY|O_BINARY);
    if(fd == -1)
        die("Can't open kernel file");
    u32 kernel_size = filelength(fd); //non-POSIX?
    if(kernel_size == -1)
        die("Can't get kernel file size");
    read_or_die(fd, rm_buf, 0x400, "Can't read first kilobyte");

    struct first1k_t* first1k = (first1k_t*)rm_buf;
    if(first1k->boot_flag != 0xAA55)
        die("No boot signature (55,AA) - it's not a kernel");
    if(first1k->header != HdrS)
        die("No 'HdrS' signature - this kernel is too old");
    if(first1k->version != 0x202)
        die("Loader protocol version is less than 2.02 - this kernel is too old");
    if(! (first1k->loadflags & 0x01))
        die("I can load bzImages high, not low, but this kernel wants low...");

    // Tell rm loader some info

    first1k->vid_mode=vid_mode;
    // last_ditch() will be called just before rm -> pm
    first1k->realmode_swtch = u32(last_ditch);
    // type of loader - unspecified: kernel do not know us (yet :-)
    first1k->type_of_loader = 0xFF;
    first1k->cmd_line_ptr = 0x98000;// 32-bit flat addr of command line
    // offset limit of the setup heap
    // heap_end_ptr appears to be relative to the start of setup (ofs 0x0200)
    first1k->loadflags |= 0x80;     // says to rm loader it's ok to use heap
    first1k->heap_end_ptr = _32k-0x200;

//TODO
//  if(loading_initrd) {
//      first1k->ramdisk_image = <initrd_address>;
//      first1k->ramdisk_size = <initrd_size>;
//  }

    // Read remaining rm loader

    // new kernels never need: if(!first1k->setup_sects) first1k->setup_sects=4;
    rm_size = 0x200*(first1k->setup_sects+1); // 0th sect is not counted there
    read_or_die(fd, rm_buf+0x400, rm_size-0x400, "Can't read rm loader");

    // Read remaining kernel (pm part)

    u16 xfer_size = _16k;
    u8* xfer_buf = malloc_or_die(xfer_size, "Can't allocate xfer buf");

    pm_size = kernel_size-rm_size;
    u32 xmm_drv = xmm_driver();
    if(xmm_drv) {
#ifdef DEBUG
        xmm_alloc(xmm_drv, 4000);   // to test XMS alloc
#endif
        pm_buf = xmm_alloc(xmm_drv, pm_size);
        if(!pm_buf)
            pm_buf = _1m + _64k; // really dangerous fallback!
            // you may place die("not enough XMM memory") here instead
            // don't forget to clean up the mess then, especially
            // XMM handle which have been lost inside xmm_alloc()...
    }
    else { // danger: no XMM, blindly using unallocated high mem
        pm_buf = _1m + _64k;    //+64k trick: keeps HMA based DOS from crashing
                                // kernel will be moved down in last_ditch()
    }

    // After XMS/raw highmem allocations above, there's no way back to DOS
    // Try to load kernel, maybe even blindly storing it in unallocated memory
    // as a last resort. Do not use exit().

    u32 high_ofs = pm_buf;
    u32 remaining = pm_size;
    while(1) {
        u16 size = read(fd, xfer_buf, xfer_size);
        if(s16(size) == -1) hang();
        if(!size) break;
        remaining -= size;
        if(s32(remaining) < 0) hang(); // too many bytes read?!!
        memcpy32(0, high_ofs, seg(xfer_buf), ofs(xfer_buf), size);
        high_ofs += size;
    }

    // replace int 3 vector: you can place int 3 (0xCC)
    // into rm loader to track how far it gets...
#ifdef DEBUG
    ((u32*)0) [3] = (u32)hang;
#endif

    // Move rm loader & commandline to 0x90000. Go!

    cli();
    memcpy32(
        0,0x90000,
        seg(rm_buf),ofs(rm_buf),
        _32k
    );
    memcpy32(
        0,0x98000,
        seg(cmdline),ofs(cmdline),
        strlen(cmdline)+1
    );
    set_sregs_jump_seg_ofs(
        0x9000,     //sregs
        0xA000,     //sp
        0x9020,0    //cs,ip
    );

    // Let compiler be happy

    return 0;
}
